/*++++++++++
.TYPE           Module
.IDENTIFICATION projection.c
.VERSION 1.0	05-Mar-1987: Creation
.VERSION 1.1	08-Sep-1988: Inclusion of several projections
.VERSION 2.0	22-Feb-1991: Tested on PCs with TuboC
.VERSION 2.1	27-Feb-1993: A few new projections (Cube)
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [ESO-IPG]  
.KEYWORDS       Projection of coordinates.
.ENV            Any
.COMMENTS
All spherical coordinates are assumed to be expressed in DEGREES.

\begin{TeX}
\def\atan{{\rm atan\/}}
\def\asin{{\rm atan\/}}
\def\acos{{\rm atan\/}}

All projections defined here have the center assumed to be at
$(\ell=0, b=0)$, i.e. $x=1$.
      The following notations are used:
\begin{itemize}
\item {\tt \_PROJ\_}... specifies the type of the projection
\item $(\ell, b)$ the longitude and latitude
\item $\vec{u} = (x, y, z)$ is unit vector (cosine direction) related to
	$(\ell, b)$ by		
	$$ \left\{ \begin{array}{rcl}
                   x & = & \cos b \cos \ell \\
                   y & = & \cos b \sin \ell \\
	                   z & = & \sin b 
		   \end{array}
	   \right. 
	$$
\item	$\theta$ is the distance from center on the celestial sphere, 
	i.e. $\theta = \cos^{-1} x$
\item 	$(X, Y)$ are the projections
\item	$\rho$ is the distance from center on the projection,
	i.e. $\rho = \sqrt{X^2 + Y^2}$
\end{itemize}

\renewcommand{\arraystretch}{2.0}

The projections are:
\large

\hspace*{-6em}\begin{tabular}{|l|ll|p{8em}|}
\hline
	{\tt \_PROJ\_} & \multicolumn{2}{|c|}{Projection formulae} & Notes \\
\hline
 {\tt TAN\_}$^{1 \dagger}$ 	& $ X = \frac{y}{x}$ 
 				& $ Y = \frac{z}{x}$ 
 				& $ \rho = \tan \theta$ \\
%\hline
 {\tt STG\_}$^2$ 		& $ X = \frac{2y}{1+x}$ 
 				& $ Y = \frac{2z}{1+x}$ 
 				& $\rho = 2 \tan\frac{\theta}{2}$ \\
%\hline
 {\tt SIN\_$^{3 \dagger}$} 	& $ X = y $ 
 				& $ Y = z $ 
 				& $\rho = \sin \theta$ \\
%\hline
 {\tt ZEA\_$^{4 \bullet}$} 	& $ X = y \sqrt{\frac{2}{1+x}}$ 
				& $ Y = z \sqrt{\frac{2}{1+x}}$ 
				& $\rho = 2 \sin\frac{\theta}{2}$ \\
%\hline
 {\tt ARC\_}$^5$ 		& $ X = y \frac{\cos^{-1}{x}}{\sqrt{y^2+z^2}}$ 
 				& $ Y = z \frac{\cos^{-1}{x}}{\sqrt{y^2+z^2}}$ 
 				& $\rho = \theta$ \\
%\hline
 {\tt GLS\_$^{6 \bullet}$} 
		& $ 	X = \ell \cos b $
		& $	Y = b $ 
		&  {\normalsize $ 	\ell\in[-\pi,+\pi]$} \\
%\hline
 {\tt MERCATOR\_} 
		& $ 	X = \ell $
		& $ 	Y = \tanh^{-1}(\sin{b}) $
		&  {\normalsize $ 	\ell\in[-\pi,+\pi]$} \\
 {\tt AITOFF\_$^\bullet$} 
		& $	X = 2\frac{\cos b\sin\frac{\ell}{2}}{\Delta} $
		& $	Y = \frac{\sin b} {\Delta} {\qquad} $ 
		&  {\normalsize $ 	\ell\in[-\pi,+\pi]$} \\
		& \multicolumn{2}{c|}{
		$\Delta = \sqrt{\frac{1+\cos b\cos\frac{\ell}{2}}{2}}$} & \\
 {\tt LAMBERT\_$^\bullet$}
 		& $ X = \ell $
 		& $ Y = \sin{b} $
		&  {\normalsize $ 	\ell\in[-\pi,+\pi]$} \\
 {\tt TSC\_$^7$} &
		\multicolumn{2}{c|}{(6 $\times$ {\tt TAN} projections)}
 		& (see {\tt TAN})	\\
 {\tt QSC\_$^{8 \bullet}$}
 		& $ X = \sqrt{\frac{\pi}{6}} . \sqrt{\frac{1-x}
 				{1-\frac{1}{\sqrt{2+(z/y)^2}}}} $
 		& $ Y/X = \frac{12}{\pi} (\atan\frac{z}{y} - 
 				\asin\frac{z}{\sqrt{2(y^2+z^2)}}) $
 		& For $\phi = \atan(Y/X) \in[-\frac{\pi}{4},+\frac{\pi}{4}]$ \\
%\hline
\hline
\end{tabular}
\normalsize
\renewcommand{\arraystretch}{1.0}
{\bf Notes}: \qquad \qquad \begin{tabular}{cl}
            $\bullet$ 	& Equal area projection \\
            $\dagger$ 	& Projection restricted to $\theta \leq \frac{\pi}{2}$ \\
            1		& Gnomonic (standard) projection \\
            2		& Stereographic projection \\
            3		& Orthographic projection \\
            4		& Zenithal Equal-Area \\
            5		& Schmidt plates projection \\
            6		& Global Sinusoidal projection \\
            7		& Tangential Spherical Cube \\
            8		& Quadrilaterized Spherical Cube \\
            \end{tabular}


\bigskip

The computations of $(x , y , z)$ from the projections
$(X , Y)$ are as follows:
\renewcommand{\arraystretch}{2.0}
\large
$$\begin{tabular}{|l|lll|}
\hline
	{\tt \_PROJ\_} & \multicolumn{3}{c|}{Unit vector components} \\
\hline
 {\tt TAN\_} 		& $ x = \frac{1}{\sqrt{1+\rho^2}}$ 
			& $ y = \frac{X}{\sqrt{1+\rho^2}}$ 
			& $ z = \frac{Y}{\sqrt{1+\rho^2}}$ \\
%\hline
 {\tt TAN2\_}	& $ x = \frac{4-\rho^2}{4+\rho^2}$ 	
	 	& $ y = \frac{4X}{4+\rho^2}$ 	
	 	& $ z = \frac{4Y}{4+\rho^2}$ 	\\
%\hline
 {\tt SIN\_}		& $ x = \sqrt{1-\rho^2} $ 
			& $ y = X $ 
			& $ z = Y $ \\
%\hline
 {\tt SIN\_2} 		& $ x = 1 - \frac{\rho^2}{2}$ 
			& $ y = X \sqrt{1-\frac{\rho^2}{4}} $
			& $ z = Y \sqrt{1-\frac{\rho^2}{4}} $ \\
%\hline
 {\tt ARC\_}		& $ x = \cos \rho $
 			& $ y = X \frac{\sin \rho}{\rho}  $
 			& $ z = Y \frac{\sin \rho}{\rho}  $  \\
%\hline
 {\tt GLS\_} 
		& $ 	x = \beta \cos \frac{X}{\beta} $
		& $ 	y = \beta \sin \frac{X}{\beta} $
		& $	z = \sin Y $ \\
   & \multicolumn{3}{|c|}{$\beta = \sqrt{1-Y^2}$}\\
%\hline
 {\tt MERCATOR\_} 
		& $ 	x = \frac{\cos X}{\cosh Y}$
		& $ 	y = \frac{\sin X}{\cosh Y}$
		& $ 	z = \tanh Y $ \\
 {\tt AITOFF\_} 	
    	& $ x = \frac{1 - \Delta^2(\frac{X^2}{2} + Y^2)}{\sqrt{1-Y^2\Delta^2}} $
    	& $ y = \frac{X\Delta(2\Delta^2-1)}{\sqrt{1-Y^2\Delta^2}}$
    	& $ z = Y \Delta $  \\
   & \multicolumn{3}{|c|}{$\Delta = \sqrt{1-\frac{X^2}{16}-\frac{Y^2}{4}}$}\\
\hline
\end{tabular}$$
\normalsize
\renewcommand{\arraystretch}{1.0}


To compute projected coordinates around 
	any $o_0 = (\ell_0, b_0)$ center:
\begin{enumerate}
\item 
	determine first the rotation matrix $R$
	with {\em tr\_oR}($o_0$, $R$)

\item	for each position $o$, 
	apply successively 
	
	{\em tr\_ou}($o$, $u$), {\em tr\_uu}($u$, $u$, $R$),
	and {\em tr\_up}({\em method}, $u$, $p$). 
	
	Conversions from projections are similarly obtained
	with successive applications of 
	
	{\em tr\_pu}({\em method}, $p$, $u$), 
	{\em tr\_uu1}($u$, $u$, $R$), and
	{\em tr\_uo}($u$, $o$).
\end{enumerate}

\end{TeX}

-----------------------------------------------------------------------*/
 
#include <stdio.h>
#include <astropos.h>
                               /* the following #define help to understand  */
#define lon          o[0] 
#define lat          o[1]
#define x            u[0]
#define y            u[1]
#define z            u[2]
#define X            p[0]
#define Y            p[1]

static short cube[7][3] = {  /* Definition of (x0,y0) and rotation for Cube */
	{ 0,0,  0 }, 	/* Face 0        */
	{ 0,2,  0 }, 	/* Face 1 (z=+1) */
	{ 2,0,  1 }, 	/* Face 2 (y=+1) */
	{ 0,0,  0 }, 	/* Face 3 (x=+1) */
	{ 4,0,  1 }, 	/* Face 4 (x=-1) */
	{-2,0,  0 }, 	/* Face 5 (y=-1) */
	{0,-2,  3 }	/* Face 6 (z=-1) */
  };

int tr_pu6(), tr_up6();

/*============================================================================*/
static int rotr(p, n)
/*++++++++++++++++++
.PURPOSE  Rotate a multiple of (pi/2)
.RETURNS  0
.METHOD	  Use unit vectors for projections.
-------------------*/
	double p[2]; 	/* MOD: Values to rotate*/
	int    n;	/* IN: Number of 90deg	*/
{
	double r;

   printf("[rot=%d]", n);
   switch(n&3) {
   	case 0:	
   		break;
   	case 1: r = Y, Y = -X, X = r;
   		break;
	case 2: X = -X, Y = -Y;
   		break;
   	case 3: r = Y, Y = X, X = -r;
   		break;
   }
   return(0);
}

static int face(p)
/*++++++++++++++++++
.PURPOSE  Compute Face Number from Projections
.RETURNS  Face Number (0 = BAD)
.METHOD	  Use unit vectors for projections.
-------------------*/
	double p[2]; 	/* IN: Projections in specified system */
{
   if ((X >= -1.0) && (X <= 1.0)) {		/* Faces 1 / 3 / 6 */
   	if (Y >   3.0)	return (0);
   	if (Y >=  1.0)	return (1);
   	if (Y >= -1.0)	return (3);
   	if (Y >= -3.0)	return (6);
   	return(0);
   }
   if ((Y > 1.0) || (Y < -1.0) || (X > 5.0))	return(0);
   if (X >=  3.0)	return(4);
   if (X >=  1.0)	return(2);
   if (X >= -3.0)	return(5);
   return (0);
}

/*============================================================================*/
int tr_op (c, o , p)
/*++++++++++++++++++
.PURPOSE  Compute projection (center is at coordinate (0,0))
.RETURNS  Projection Number (0 = BAD)
.METHOD	  Use unit vectors for projections.
-------------------*/
	int	c;	/* IN: The projection system	*/
	double o[2]; 	/* IN: Polar angles to project */
	double p[2]; 	/* OUT: Projections in specified system */
{
	double us[3];

  tr_ou(o, us);
  return(tr_up(c, us, p));
}

/*============================================================================*/
int tr_po ( c, p , o )   
/*++++++++++++++
.PURPOSE  Derive coordinates from a projection centered at position (0,0).
.METHOD   unit vectors are used.
.RETURNS  0 if bad projection system or point outside limits,
	  projection number otherwise
--------------*/
	int	c;	/* IN: The projection system	*/
	double p[2]; 	/* IN: Projections in specified system	*/
	double o[2]; 	/* OUT: Polar angles         */
{ 
	double 	us[3];
	int	ret;

  if ((ret = tr_pu(c, p, us)))  tr_uo(us, o);
  return(ret);
}

/*============================================================================*/
int tr_pu ( c, p , u )  
/*+++++++++
.PURPOSE  Computes Unit vector from a position in projection
		centered at position (0,0).
.RETURNS  0 if the point is outside the limits, or if the projection is unknown.
	  A non-null number for success (face number for cube projection)
---------*/
	int	c;	/* IN: The projection system	*/
	double p[2]; 	/* IN: Projection in `c' system */
	double u[3]; 	/* OUT: unit vector          */
{
	double 	r,s, pp[2];
	int	ret, k;

  switch(ret = c) {
    default:
	ret = 0;
	break;

    case _PROJ_AITOFF_:		/* Limit is ellipse with axises 
    				 * a = 2 * sqrt(2) ,  b = sqrt(2)  */
                           		/*     Compute dir l/2, b  */
	r = X*X/8.e0 + Y*Y/2.e0; 	/* 1 - cos b . cos l/2 */
	if (r > 1.e0)  			/* Test outside domain */
     		{ ret = 0; break; }
	x = 1.e0 - r ;   		/*     cos b . cos l/2 */
	s = sqrt(1.e0 - r/2.e0) ;   	/* sqrt(( 1 + cos b . cos l/2)/2) */
	y = X * s / 2.e0;
	z = Y * s ;
					/* From (l/2,b) to (l,b)   	*/
	r = hypot ( x, y ) ;		/* cos b			*/
	if (r)	
	{	s = x;
		x = (s*s - y*y) /r;
		y = 2.e0 * s * y/r;
	}
	break;

    case _PROJ_GLS_:		/* Limit is |Y| <= pi/2 */
	z = sin(Y);
	r = 1 - z*z;		/* cos(b) ** 2	*/
	if (r < 0.0e0)
		{ ret = 0; break; }
	r = sqrt(r);		/* cos b	*/
	if (r)	s = X/r;	/* Longitude	*/
	else	s = 0.e0;	/* For poles	*/
	x = r * cos(s);
	y = r * sin(s);
	break;

    case _PROJ_MERCATOR_:
	z = tanh(Y), r = 1.e0/cosh(Y);
	x = r * cos(X);
	y = r * sin(X);
	break;

    case _PROJ_LAMBERT_:	/* Always possible */
    	z = Y;
	r = 1 - z*z;		/* cos(b) ** 2	*/
	if (r < 0.0e0)
		{ ret = 0; break; }
	r = sqrt(r);		/* cos b	*/
	x = r * cos(X);
	y = r * sin(X);
	break;
	
    case _PROJ_TAN_:		/* No limit			*/
	x = 1.e0 / sqrt(1.e0 + X*X + Y*Y);
	y = X * x;
	z = Y * x;
	break;

    case _PROJ_TAN2_:		/* No limit			*/
	r = (X*X + Y*Y)/4.e0;
	s = 1.e0 + r;
	x = (1.e0 - r)/s;
	y = X / s;
	z = Y / s;
	break;

    case _PROJ_ARC_:		/* Limit is circle, radius PI	*/
	r = hypot(X, Y);
	if (r > PI)	
		{ ret = 0; break; }
	s = sinc(r);
	x = cos(r);
	y = s * X;
	z = s * Y;
	break;

    case _PROJ_SIN_:		/* Limit is circle, radius 1	*/
	s = 1.e0 - X*X - Y*Y;
	if (s < 0.e0)
		{ ret = 0; break; }
	x = sqrt(s);
	y = X;
	z = Y;
	break;

    case _PROJ_SIN2_:		/* Limit is circle, radius 2	*/
	r = (X*X + Y*Y)/4.e0;
	if (r > 1.e0)
		{ ret = 0; break; }
	s = sqrt(1.e0 - r);
	x = 1.e0 - 2.e0 * r;
	y = s * X;
	z = s * Y;
	break;

    case _PROJ_TSC_:		/* Spherical Cube		*/ 
    	ret = face(p);
    	printf(".... Face: %d", ret);
    	if (!ret) 	break;		/* ... Outside limits	*/
    	pp[0] = p[0] - cube[ret][0];
    	pp[1] = p[1] - cube[ret][1];
    	printf("; %11.8f,%11.8f", pp[0], pp[1]);
    	rotr(pp, 4-cube[ret][2]);
    	printf("; %11.8f,%11.8f\n", pp[0], pp[1]);
    	tr_pu6(pp, u, ret);
    	break;
    	
    case _PROJ_QSC_:		/* Spherical Cube, Equal-Area	*/ 
    	r = sqrt(PI/6.0);
    	pp[0] = p[0] / r; pp[1] = p[1] / r;
    	ret = face(pp);
    	printf(".... Face: %d", ret);
    	if (!ret) 	break;		/* ... Outside limits	*/
    	pp[0] -= cube[ret][0];
    	pp[1] -= cube[ret][1];
    	printf(" (%11.8f,%11.8f)", pp[0], pp[1]);
    	rotr(pp, 4-cube[ret][2]);
    	printf(" (%11.8f,%11.8f)\n", pp[0], pp[1]);
    	r = fabs(pp[1]);
    	if (pp[0] >= r) k = 0;
    	else if (pp[0] <= -r) k = 2;
    	else if (pp[1] >= 0)  k = 1;
    	else k = 3;
    	rotr(pp, k);
   	printf(" Quad=(%11.8f,%11.8f)", pp[0], pp[1]);
    	if (pp[0] > 0.) {	double sm1;
    		s = PI*pp[1]/(12.*pp[0]);
    		r = sin(s)/(cos(s) - 1./sqrt(2.));	/* = Y/X */
    		s = pp[0]*pp[0]*(1.-1./sqrt(2.+r*r));	/* 1-cos(b) */
    		sm1 = 1 - s;
    		pp[0] = s * (2.-s)/(sm1*sm1*(1+r*r));
    		pp[0] = sqrt(pp[0]); 
    		pp[1] = pp[0] * r;
 	   	printf(" Gno=(%11.8f,%11.8f)", pp[0], pp[1]);
	}
    	rotr(pp, 4-k);
    	printf("; %11.8f,%11.8f\n", pp[0], pp[1]);
    	tr_pu6(pp, u, ret);
    	break;
  }

  if (ret == 0)	x = y = z = 0.0;

  return(ret);
}

/*============================================================================*/
int tr_up(c, u , p ) 
/*++++++++++++
.PURPOSE Compute projections from unit vector.
\begin{TeX}
The center of the projection corresponds to $\vec{u} = (1, 0, 0)$.
\end{TeX}
.RETURNS 0 if u == 0 (not a unit vector)
         Projection system otherwise
-------------*/
	int	c;	/* IN: The projection system	*/
	double u[3]; 	/* IN: Unit vector */
	double p[2]; 	/* OUT: Projections */
{
	double  r, den, pp[2];
	int	ret;	/* Status */
	int	k;

#if DEBUG
  printf("tr_up: proj.%d -- direction=%13.9f,%13.9f,%13.9f\n",c,x,y,z);
#endif

  r = hypot(x,y);			/* r = cos b	*/
  if ( (r == 0.e0) && (z == 0.e0))	return(0);

  switch(ret = c) {
    default:
  	ret = 0;
  	break;

    case _PROJ_AITOFF_:
	den = sqrt (r*(r+x)/2.e0); 		/*   cos b . cos l/2 */
	X = sqrt(2.*r*(r-x));
	den = sqrt ((1.e0 + den)/2.e0); 
	X   = X / den;
	Y   = z / den;
      	if (y < 0.e0)	X = -X;
	break;

    case _PROJ_GLS_:
    	Y = asin(z);				/* sin b 	*/
	if (r)	X = atan2(y,x) * r;		/* l cos b 	*/
	else	X = 0.e0;
	break;

    case _PROJ_MERCATOR_:
	if (r)	X = atan2(y,x), Y = atanh(z);
	else	ret = 0;
	break;

    case _PROJ_TAN_:
	if (x > 0.e0)	X = y/x, Y = z/x; 
	else		ret = 0;
	break;

    case _PROJ_TAN2_:
    	den = (1.e0 + x)/2.e0;
    	if (den > 0.e0)	X = y/den, Y = z/den;
    	else		ret = 0;
    	break;

    case _PROJ_ARC_:
	if (x <= -1.e0)			/* Distance of 180 degrees */
		X = PI, Y = 0.e0;
	else				/* Arccos(x) = Arcsin(r)   */
	{	r = hypot(y,z);
		if (x > 0.e0)	den = asinc(r);
		else		den = acos(x)/r;
		X = y * den;
		Y = z * den;
	}
    	break;

    case _PROJ_SIN_:
	if (x >= 0.e0)	X = y, Y = z; 
	else		ret = 0;
	break;

    case _PROJ_SIN2_:	/* Always possible */
	den = sqrt((1.e0 + x)/2.e0);
	if (den)	X = y / den, Y = z / den;
	else		X = 2.e0, Y = 0.e0;	/* For x = -1 */
	break;

    case _PROJ_LAMBERT_:	/* Always possible */
     	Y = z;	X = 0;
	if (r)	X = atan2(y,x); 
	break;

    case _PROJ_TSC_:	/* Always possible */
    	ret = tr_up6(u, p, 0);	/* Projection on one face */
    	printf(".... Face: %d (%11.8f,%11.8f)", ret, p[0], p[1]);
    	rotr(p, cube[ret][2]);
    	printf("(%11.8f,%11.8f)", p[0], p[1]);
    	printf(" Off: %d %d\n", cube[ret][0], cube[ret][1]);
    	X += cube[ret][0];
    	Y += cube[ret][1];
    	break;

    case _PROJ_QSC_:	/* Always possible */
    	ret = tr_up6(u, p, 0);	/* Projection on one face */
    	printf(".... Face: %d (%11.8f,%11.8f)", ret, p[0], p[1]);
    	r = fabs(Y);
    	if (X >= r) k = 0;
    	else if (X <= -r) k = 2;
    	else if (Y >= 0)  k = 1;
    	else k = 3;
    	rotr(p, k);
    	printf(". Rot(%d)=(%11.8f,%11.8f)\n", k, p[0], p[1]);
    	if (X > 0.) {	
    		den = X*X + Y*Y; r = Y/X;
    		pp[0] = sqrt((1.-1./sqrt(1.+den))/(1.-1./sqrt(2.0+r*r)));
    		pp[1] = (12.0/PI)*(atan(r) - asin(Y/(sqrt(2.0*den))));
    		pp[1] *= pp[0];
    		X = pp[0]; Y = pp[1];
 	   	printf(".... Quad=   (%11.8f,%11.8f)", p[0], p[1]);
	}
    	rotr(p, 4-k);
    	printf(". Rot(%d)=(%11.8f,%11.8f)\n", k, p[0], p[1]);
    	rotr(p, cube[ret][2]);
    	printf("    =>(%11.8f,%11.8f)", p[0], p[1]);
    	printf(" Off: %d %d\n", cube[ret][0], cube[ret][1]);
    	X += cube[ret][0];
    	Y += cube[ret][1];
    	printf(" Cube(%11.8f,%11.8f)\n", p[0], p[1]);
    	r = sqrt(PI/6.);
    	X *= r; Y *= r;
    	break;
  }

  return(ret);
}
