/*++++++++++++++
.IDENTIFICATION compute_qbox.c
.LANGUAGE       C
.AUTHOR         CDS Catalogues Service
.ENVIRONMENT    
.KEYWORDS       
.VERSION  1.0   31-Dec-1998
.VERSION  1.1   05-Jun-2003: New DENIS Version
.VERSION  1.2   01-Mar-2007: SDSS
.VERSION  1.3   22-Feb-2012: Main program
.COMMENTS       Compute Qboxes for a catalog.
	(Input = RA / Dec in degrees)
---------------*/

#define VERSION "1.2"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <qbox.h>
#include <math.h>

#define QBOX_LEVEL 8
#define QBOX_SIZE  (6<<(QBOX_LEVEL<<1))
#ifndef int8
#define int8 long long
#endif

double atof() ;
static double test_pos[2] = { 153.6545 , 56.8091 };
static double max_rad = 5./3600. ;
static int verbop = 0;
static char usage[] = "\
Usage: qboxes [-version] [-b binfile_name] [file...|-]\n";
static char help[] = "\
 -version: print the version & exit\n\
       -b: save a binary version of qboxes\n\
     file: file with position (in degrees); use - for stdin\n\
The positions must be in the 2 leftmost columns\n\
";

int main(int argc, char **argv)
{
  char buf[1024] ;
  static char *binfile = 0;
  static int level = QBOX_LEVEL;
  int *counts, *u, *e, *c ;
  char *p, *in_test, *filename;
  double pos[2] ;
  int i, n, lev, rec=0, recno=0, nf=0;
  long long bytes = 0;
  FILE *file;

    /* Initialize Qboxes */
    qbox_set(level /*QBOX_LEVEL*/) ;
    in_test = malloc(QBOX_SIZE); memset(in_test, 0, QBOX_SIZE);
    counts = malloc(i = QBOX_SIZE*sizeof(int)) ;
    memset(counts, 0, i) ;

    /* Arguments */
    if (argc<2) {
	fprintf(stderr, "%s", usage);
	exit(1);
    }
    while (--argc>=0) {
	p = *++argv;
	if (p && (*p == '-')) switch(p[1]) {
	  case 0:	/* - alone ==> positions in stdin */
	    break;
	  case 'h':	/* Assume help */
	    printf("%s%s", usage, help);
	    exit(0);
#if 0
	  case 'l':	/* Assume level */
	    while (isalpha(*p) || (*p == '=')) p++;
	    if (isgraph(*p)) level = atoi(p);
	    continue;
#endif
	  case 'b':
	    argc--; binfile = *++argv;
	    continue;
	  case 'v':
	    if (strncmp(p, "-ver", 4) == 0) {
		printf("compute_qbox (CDS) Version %s\n", VERSION);
		exit(0);
	    }
	    /* Assume verbose -- ignored in fact! */
	    verbop = 1;
	    continue;
	  default:
	    fprintf(stderr, "#***Unknown option: %s\n%s\n", p, usage);
	    exit(1);
	}
	if (!p) {	/* Last argument */
	    if (nf) break;
	    p = "-";
	}
	filename = p;
	file = (p[0]=='-')&&(p[1]==0) ? stdin : fopen(p, "r");
	if (!file) {
	    fprintf(stderr, "#*** file %s: ", filename);
	    perror("");
	    exit(2);
	}
	nf++;		/* Counter of files */
    
        /* Read the list of Strips, and compute the Qboxes */
        while(fgets(buf, sizeof(buf), file)) {
	    bytes += strlen(buf); recno++;
	    for (p=buf; *p == ' '; p++) ;
	    if (!isdigit(*p)) { 	/* Not a position */
		if (verbop) fprintf(stderr, 
			"#...ignored line#%d: %s", recno, buf);
	        continue;
	    }
	    rec++;
	    pos[0] = atof(p);		/* RA */
	    while (isdigit(*p) || (*p == '.')) p++ ; 
	    while (isspace(*p)) p++ ;
	    if ((p[0]=='+') || (p[0]=='-') || isdigit(p[0]))
	        pos[1] = atof(p); 		/* Dec */
	    else {
		fprintf(stderr, "#***bad declination [%s]: %s", 
			filename, buf);
		rec--;
		continue;
	    }
	    i = qbox_o(pos) - (9<<(QBOX_LEVEL<<1)) ;
	    counts[i] += 1 ;
	    if (dist_o(pos, test_pos) <= max_rad) {
	        fprintf(stderr, 
	 	    "#...in test(%09.5f%+09.5f): %09.5f%+09.5f #%06d\n",
		    test_pos[0], test_pos[1], pos[0], pos[1], i);
	        in_test[i] |= 1;
	    }
        }
        fprintf(stderr, "#===Input: %d positions, bytes=%lld\n", rec, bytes);
	if (file != stdin) fclose(file);
    }

    /* Copy the binary version */
    if (binfile) {
	lev = open(binfile, 1, 0644);
        if (lev<0) { 
	    fprintf(stderr, "#***write binary qboxes: %s: ", binfile);
	    perror("");
	    exit(2);
	}
        write(lev, counts, QBOX_SIZE*sizeof(int));
	if (verbop) fprintf(stderr, 
		"#---written binary qboxes (%d bytes): %s\n",
		(int)(QBOX_SIZE*sizeof(int)), binfile);
        close(lev);
    }

    printf("#---Total: ");
    for (i=n=lev=0; i<QBOX_SIZE; i++) {
	if (counts[i]) lev++;
	n += counts[i];
    }
    printf("%d non-empty boxes for %d stars\n", lev, n);

    /* Group the boxes */
    for (lev = QBOX_LEVEL; lev > 0; lev--) {
	i = 9<<(lev<<1) ;
	c = u = counts ;
	e = c + (6<<(lev<<1)) ;
	while (c<e) {
	    if (c[0] && c[1] && c[2] && c[3]) {
		*u = (c[0] + c[1] + c[2] + c[3])>>2;
		c += 4;
		u++;
		continue ;
	    }
	    if (c[0]) printf("%d #d=%d\n", i+(int)(c-counts), *c); c++ ;
	    if (c[0]) printf("%d #d=%d\n", i+(int)(c-counts), *c); c++ ;
	    if (c[0]) printf("%d #d=%d\n", i+(int)(c-counts), *c); c++ ;
	    if (c[0]) printf("%d #d=%d\n", i+(int)(c-counts), *c); c++ ;
	    *u = 0;
	    u++ ;
	}
    }
    /* Write out the list of Qboxes */
    for (i=0; i<6; i++) {
        if (counts[i]) printf("%d\n", 9+i) ;
    }
    exit(0) ;
}

