/*+++++++++++++++       
.TYPE           Module
.IDENTIFICATION supergal.c
.VERSION 1.0	27-Feb-1987: Creation .
.VERSION 1.1	27-Oct-1988: Use Unit Vectors
.VERSION 2.0	22-Feb-1991: Tested on PCs with TuboC
.LANGUAGE       C
.AUTHOR         Francois Ochsenbein [ESO-IPG]  
.KEYWORDS       Galactic / Supergalactic conversions

.ENV            

.COMMENTS
All spherical coordinates are assumed to be expressed in DEGREES.
No function is traced.

\begin{TeX}

The SuperGalactic system is defined by the positions of
$$\begin{tabular}{|rr|rr|} \hline
	$L$ & $B$ & $\ell$ & $b$ \\ \hline
	 $0\d$ & $+90\d$ & $47.37\d$ & $+6.32\d$ \\
	$0\d$  & $0\d$   & $137.37\d$ & $0.00\d$ \\
	\hline
\end{tabular}$$


The parameter mnemonics are:

\begin{itemize}
 \item g = array $[\ell,b]$    of  Galactic   coordinates
 \item G = array $[L,B]$ of SuperGalactic coordinates.
\end{itemize}
\end{TeX}

-----------------------------------------------------------------------*/
 

#include <astropos.h>
#include <trigo.h>

double supergal[3][3] = {	/* Galactic to SuperGalactic	*/
	{-0.7357425748043749, 0.6772612964138942,        0.},
	{-0.0745537783652337,-0.0809914713069767, 0.9939225903997749},
	{ 0.6731453021092076, 0.7312711658169645, 0.1100812622247821}
			};

/*============================================================================*/
int tr_gsup (g , G)  
/*+++
.PURPOSE   transformation from Galactic to SuperGalactic
.RETURNS   1
---*/
	double g[3]; 	/* IN: galactic angles l and b in degrees */
	double G[3]; 	/* OUT: Supergalactic angles in degrees */
{ 
  return (tr_uu ( g, G, supergal));
}


/*============================================================================*/
int tr_supg (G , g)  
/*+++
.PURPOSE transformation from SuperGalactic to Galactic 
.RETURNS 1
---*/
	double G[3];	/* IN: Supergalactic angles in degrees */
	double g[3]; 	/* OUT: galactic angles l and b */
{ 
  return (tr_uu1 ( G, g, supergal));
}

